#include "aron_conversions.h"

// STL
#include <stdexcept>
#include <string>

// Ice
#include <IceUtil/Time.h>

// Simox
#include <SimoxUtility/shapes/OrientedBox.h>

// RobotAPI
// TODO: ice dependency! header should be removed.
#include <RobotAPI/interface/objectpose/object_pose_types.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectPose.h>
#include <RobotAPI/libraries/ArmarXObjects/aron/ObjectPose.aron.generated.h>


namespace armarx::objpose
{

    ObjectID fromAron(const aron::ObjectID& id)
    {
        return ObjectID(id.dataset, id.className, id.instanceName);
    }

    ObjectTypeEnum fromAron(const aron::ObjectTypes& objectType)
    {
        using AronObjectType = aron::ObjectTypes::__ImplEnum;

        ObjectTypeEnum e{};

        switch (objectType.value)
        {
            case AronObjectType::ANY_OBJECT:
                e = ObjectTypeEnum::AnyObject;
                break;
            case AronObjectType::KNOWN_OBJECT:
                e = ObjectTypeEnum::KnownObject;
                break;
            case AronObjectType::UNKNOWN_OBJECT:
                e = ObjectTypeEnum::UnknownObject;
                break;
        }

        return e;
    }

    simox::OrientedBoxf fromAron(const aron::OrientedBoundingBox& obb)
    {
        return simox::OrientedBoxf(obb.centerPose, obb.extends);
    }

    ObjectAttachmentInfo fromAron(const aron::ObjectAttachmentInfo& info)
    {
        return
        {
            .frameName = info.frameName,
            .agentName = info.agentName,
            .poseInFrame = info.poseInFrame
        };
    }


    void fromAron(const aron::ObjectPose& aronObjectPose, ObjectPose& objectPose)
    {
        objectPose.providerName = aronObjectPose.providerName;

        objectPose.objectType = fromAron(aronObjectPose.objectType);

        objectPose.objectID = fromAron(aronObjectPose.objectID);

        objectPose.objectPoseRobot = aronObjectPose.objectPoseRobot;
        objectPose.objectPoseGlobal = aronObjectPose.objectPoseGlobal;
        objectPose.objectPoseOriginal = aronObjectPose.objectPoseOriginal;
        objectPose.objectPoseOriginalFrame = aronObjectPose.objectPoseOriginalFrame;

        objectPose.robotConfig = aronObjectPose.robotConfig;
        objectPose.robotPose = aronObjectPose.robotPose;

        objectPose.attachment = fromAron(aronObjectPose.attachment);

        objectPose.confidence = aronObjectPose.confidence;

        objectPose.timestamp = IceUtil::Time::microSeconds(aronObjectPose.timestamp);

        objectPose.localOOBB = fromAron(aronObjectPose.localOOBB);

    }


    aron::ObjectID toAron(const ObjectID& id)
    {
        aron::ObjectID aronId;

        aronId.className = id.className();
        aronId.dataset = id.dataset();
        aronId.instanceName = id.instanceName();

        return aronId;
    }

    aron::ObjectTypes toAron(const ObjectTypeEnum& objectType)
    {
        aron::ObjectTypes ot{};

        switch (objectType)
        {
            case ObjectTypeEnum::AnyObject:
                ot = aron::ObjectTypes::ANY_OBJECT;
                break;
            case ObjectTypeEnum::KnownObject:
                ot = aron::ObjectTypes::KNOWN_OBJECT;
                break;
            case ObjectTypeEnum::UnknownObject:
                ot = aron::ObjectTypes::UNKNOWN_OBJECT;
                break;
        }

        return ot;
    }

    aron::ObjectAttachmentInfo toAron(const ObjectAttachmentInfo& info)
    {
        aron::ObjectAttachmentInfo aronInfo;

        aronInfo.agentName = info.agentName;
        aronInfo.frameName = info.frameName;
        aronInfo.poseInFrame = info.poseInFrame;

        return aronInfo;
    }

    aron::OrientedBoundingBox toAron(const simox::OrientedBoxf& box)
    {
        aron::OrientedBoundingBox aronBox;
        aronBox.centerPose = box.transformation();
        aronBox.extends = box.dimensions();
        return aronBox;
    }

    template <typename T>
    auto toAron(const std::optional<T>& p)
    {
        if (p)
        {
            return toAron(p.value());
        }

        // TODO(fabian.reister): handle optional properly
        // current fallback: default c'tor
        return decltype(toAron(p.value()))();
    }


    void toAron(const ObjectPose& objectPose, aron::ObjectPose& aronObjectPose)
    {
        aronObjectPose.providerName = objectPose.providerName;

        aronObjectPose.objectType = toAron(objectPose.objectType);

        aronObjectPose.objectID = toAron(objectPose.objectID);

        aronObjectPose.objectPoseRobot = objectPose.objectPoseRobot;
        aronObjectPose.objectPoseGlobal = objectPose.objectPoseGlobal;
        aronObjectPose.objectPoseOriginal = objectPose.objectPoseOriginal;
        aronObjectPose.objectPoseOriginalFrame = objectPose.objectPoseOriginalFrame;

        aronObjectPose.robotConfig = objectPose.robotConfig;
        aronObjectPose.robotPose = objectPose.robotPose;

        aronObjectPose.attachment = toAron(objectPose.attachment);

        aronObjectPose.confidence = objectPose.confidence;

        aronObjectPose.timestamp = objectPose.timestamp.toMicroSeconds();

        aronObjectPose.localOOBB = toAron(objectPose.localOOBB);
    }

} // namespace armarx::objpose
