#include "Entity.h"

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem
{

    Entity::Entity()
    {
    }

    Entity::Entity(const std::string& name, const MemoryID& parentID) :
        Entity(parentID.withEntityName(name))
    {
    }

    Entity::Entity(const MemoryID& id) : Base(id)
    {
    }

    Entity::Entity(const Entity& other) : Base(other)
    {
        *this = other;
    }

    Entity& Entity::operator=(const Entity& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    bool Entity::equalsDeep(const Entity& other) const
    {
        if (size() != other.size())
        {
            return false;
        }
        for (const auto& [key, snapshot] : history)
        {
            if (not other.hasSnapshot(key))
            {
                return false;
            }
            if (not snapshot.equalsDeep(other.getSnapshot(key)))
            {
                return false;
            }
        }
        return true;
    }


    void Entity::_copySelf(Entity& other) const
    {
        Base::_copySelf(other);
        other.maxHistorySize = this->maxHistorySize;
    }

    void Entity::_copyElements(Entity& other, bool withData) const
    {
        other.history.clear();
        for (const auto& [time, snapshot] : this->history)
        {
            other.addSnapshot(snapshot.copy(withData));
        }
    }


    bool Entity::hasSnapshot(Time time) const
    {
        return history.count(time) > 0;
    }

    bool Entity::hasSnapshotData(Time time) const
    {
        return history.count(time) > 0;
    }

    EntitySnapshot& Entity::getSnapshot(Time time)
    {
        return const_cast<EntitySnapshot&>(const_cast<const Entity*>(this)->getSnapshot(time));
    }

    const EntitySnapshot& Entity::getSnapshot(Time time) const
    {
        auto it = history.find(time);
        if (it != history.end())
        {
            return it->second;
        }
        else
        {
            throw error::MissingEntry("entity snapshot", toDateTimeMilliSeconds(time),
                                      getLevelName(), this->id().str());
        }
    }

    EntitySnapshot& Entity::getSnapshot(const MemoryID& id)
    {
        return const_cast<EntitySnapshot&>(const_cast<const Entity*>(this)->getSnapshot(id));
    }

    const EntitySnapshot& Entity::getSnapshot(const MemoryID& id) const
    {
        checkEntityName(id.entityName);
        return getSnapshot(id.timestamp);
    }

    EntitySnapshot& Entity::getLatestSnapshot()
    {
        return const_cast<EntitySnapshot&>(const_cast<const Entity*>(this)->getLatestSnapshot());
    }

    const EntitySnapshot& Entity::getLatestSnapshot() const
    {
        return getLatestItem().second;
    }

    Time Entity::getLatestTimestamp() const
    {
        return getLatestItem().first;
    }

    std::vector<Time> Entity::getTimestamps() const
    {
        return simox::alg::get_keys(history);
    }


    EntitySnapshot& Entity::addSnapshot(const EntitySnapshot& snapshot)
    {
        return addSnapshot(EntitySnapshot(snapshot));
    }


    EntitySnapshot& Entity::addSnapshot(EntitySnapshot&& snapshot)
    {
        auto it = history.emplace(snapshot.time(), std::move(snapshot)).first;
        it->second.id().setEntityID(id());
        return it->second;
    }

    EntitySnapshot& Entity::addSnapshot(const Time& timestamp)
    {
        return addSnapshot(EntitySnapshot(timestamp));
    }

    MemoryID Entity::update(const EntityUpdate& update)
    {
        checkEntityName(update.entityID.entityName);
        id() = update.entityID;

        EntitySnapshot* snapshot;

        auto it = history.find(update.timeCreated);
        if (it == history.end())
        {
            // Insert into history.
            snapshot = &addSnapshot(update.timeCreated);
            truncateHistoryToSize();
        }
        else
        {
            snapshot = &it->second;
        }
        // Update entry.
        snapshot->update(update);

        return snapshot->id();
    }


    void Entity::setMaxHistorySize(long maxSize)
    {
        this->maxHistorySize = maxSize;
        truncateHistoryToSize();
    }


    void Entity::truncateHistoryToSize()
    {
        if (maxHistorySize >= 0)
        {
            while (history.size() > size_t(maxHistorySize))
            {
                history.erase(history.begin());
            }
            ARMARX_CHECK_LESS_EQUAL(history.size(), maxHistorySize);
        }
    }

    void Entity::checkEntityName(const std::string& name) const
    {
        if (name != this->name())
        {
            throw armem::error::ContainerNameMismatch(name, getLevelName(), this->name());
        }
    }

    const std::map<Time, EntitySnapshot>::value_type& Entity::getLatestItem() const
    {
        if (history.empty())
        {
            throw armem::error::EntityHistoryEmpty(name(), "when getting the latest snapshot.");
        }
        return *history.rbegin();
    }

}
