#include "EntitySnapshot.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "error.h"


namespace armarx::armem
{


    EntitySnapshot::EntitySnapshot()
    {
    }

    EntitySnapshot::EntitySnapshot(Time time, const MemoryID& parentID) :
        EntitySnapshot(parentID.withTimestamp(time))
    {
    }

    EntitySnapshot::EntitySnapshot(const MemoryID& id) : Base(id)
    {
    }

    EntitySnapshot::EntitySnapshot(const EntitySnapshot& other) : Base(other)
    {
        *this = other;
    }

    EntitySnapshot& EntitySnapshot::operator=(const EntitySnapshot& other)
    {
        other._copySelf(*this);
        other._copyElements(*this, true);
        return *this;
    }

    bool EntitySnapshot::equalsDeep(const EntitySnapshot& other) const
    {
        if (size() != other.size())
        {
            return false;
        }
        int i = 0;
        for (const auto& instance : instances)
        {
            if (not instance.equalsDeep(other.getInstance(i)))
            {
                return false;
            }
            i++;
        }
        return true;
    }


    void EntitySnapshot::_copyElements(EntitySnapshot& other, bool withData) const
    {
        other.instances.clear();
        other.instances.reserve(this->instances.size());
        for (const auto& data : this->instances)
        {
            other.instances.emplace_back(data.copy(withData));
        }
    }


    void EntitySnapshot::update(const EntityUpdate& update,
                                std::optional<MemoryID> parentID)
    {
        if (parentID)
        {
            id() = *parentID;
        }
        time() = update.timeCreated;

        instances.clear();
        for (int i = 0; i < int(update.instancesData.size()); ++i)
        {
            EntityInstance& data = instances.emplace_back(i, id());
            data.update(update, i);
        }
    }

    bool EntitySnapshot::hasInstance(int index) const
    {
        size_t si = size_t(index);
        return index >= 0 && si < instances.size();
    }


    EntityInstance& EntitySnapshot::getInstance(int index)
    {
        return const_cast<EntityInstance&>(const_cast<const EntitySnapshot*>(this)->getInstance(index));
    }

    const EntityInstance& EntitySnapshot::getInstance(int index) const
    {
        if (hasInstance(index))
        {
            size_t si = size_t(index);
            return instances[si];
        }
        else
        {
            throw armem::error::MissingEntry(EntityInstance().getLevelName(), std::to_string(index),
                                             getLevelName(), toDateTimeMilliSeconds(time()));
        }
    }


    EntityInstance& EntitySnapshot::getInstance(const MemoryID& id)
    {
        return const_cast<EntityInstance&>(const_cast<const EntitySnapshot*>(this)->getInstance(id));
    }

    const EntityInstance& EntitySnapshot::getInstance(const MemoryID& id) const
    {
        if (!id.hasInstanceIndex())
        {
            throw armem::error::InvalidMemoryID(id, "ID has no instance index.");
        }
        return getInstance(id.instanceIndex);
    }

    EntityInstance& EntitySnapshot::addInstance(const EntityInstance& instance)
    {
        return addInstance(EntityInstance(instance));
    }


    EntityInstance& EntitySnapshot::addInstance(EntityInstance&& instance)
    {
        if (instance.index() > 0 && (size_t) instance.index() < instances.size())
        {
            throw error::InvalidArgument(std::to_string(instance.index()), "EntitySnapshot::addInstance",
                                         "Cannot add an EntityInstance because its index already exists.");
        }
        if (instance.index() > 0 && (size_t) instance.index() > instances.size())
        {
            throw error::InvalidArgument(std::to_string(instance.index()), "EntitySnapshot::addInstance",
                                         "Cannot add an EntityInstance because its index is too big.");
        }

        int new_index = instances.size();
        auto& it = instances.emplace_back(std::move(instance));
        it.index() = new_index;
        return it;
    }

    std::string EntitySnapshot::getKeyString() const
    {
        return toDateTimeMilliSeconds(this->time());
    }

    std::string EntitySnapshot::getLevelName() const
    {
        return "entity snapshot";
    }

    EntitySnapshot::ContainerT& EntitySnapshot::_underlyingContainer()
    {
        return instances;
    }

}
