#include "TreeWidget.h"

#include <RobotAPI/libraries/aron/core/navigator/type/container/Object.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <SimoxUtility/algorithm/string.h>

#include <QHeaderView>


namespace armarx::armem::gui::memory
{

    TreeWidget::TreeWidget()
    {
        initBuilders();
        initWidget();
    }

    void TreeWidget::initWidget()
    {
        clear();
        QStringList columns;
        columns.insert(int(Columns::KEY), "Key");
        columns.insert(int(Columns::SIZE), "#");
        columns.insert(int(Columns::TYPE), "Type");
        columns.insert(int(Columns::LEVEL), "Level");
        columns.insert(int(Columns::ID), "Memory ID");
        setColumnCount(columns.size());
        setHeaderLabels(columns);

        header()->setMinimumSectionSize(25);
        header()->resizeSection(int(Columns::KEY), 250);
        header()->resizeSection(int(Columns::SIZE), 25);
        header()->setTextElideMode(Qt::TextElideMode::ElideRight);


        connect(this, &This::updated, this, &This::handleSelection);
        connect(this, &QTreeWidget::currentItemChanged, this, &This::handleSelection);

        connect(this, &This::memorySelected, this, &This::itemSelected);
        connect(this, &This::coreSegmentSelected, this, &This::itemSelected);
        connect(this, &This::providerSegmentSelected, this, &This::itemSelected);
        connect(this, &This::entitySelected, this, &This::itemSelected);
        connect(this, &This::snapshotSelected, this, &This::itemSelected);
        connect(this, &This::instanceSelected, this, &This::itemSelected);
    }

    void TreeWidget::initBuilders()
    {
        memoryBuilder.setExpand(true);
        memoryBuilder.setMakeItemFn([this](const std::string & name, const Memory * memory)
        {
            return makeItem(name, *memory);
        });
        memoryBuilder.setUpdateItemFn([this](const std::string&, const Memory * memory, QTreeWidgetItem * memoryItem)
        {
            updateContainerItem(*memory, memoryItem);
            if (memoryItem)
            {
                updateChildren(*memory, memoryItem);
            }
            return true;
        });


        coreSegmentBuilder.setExpand(true);
        coreSegmentBuilder.setMakeItemFn([this](const std::string & name, const CoreSegment & coreSeg)
        {
            return makeItem(name, coreSeg);
        });
        coreSegmentBuilder.setUpdateItemFn([this](const std::string&, const CoreSegment & coreSeg, QTreeWidgetItem * coreSegItem)
        {
            updateTypedContainerItem(coreSeg, coreSegItem);
            updateChildren(coreSeg, coreSegItem);
            return true;
        });

        provSegmentBuilder.setExpand(true);
        provSegmentBuilder.setMakeItemFn([this](const std::string & name, const ProviderSegment & provSeg)
        {
            return makeItem(name, provSeg);
        });
        provSegmentBuilder.setUpdateItemFn([this](const std::string&, const ProviderSegment & provSeg, QTreeWidgetItem * provSegItem)
        {
            updateTypedContainerItem(provSeg, provSegItem);
            updateChildren(provSeg, provSegItem);
            return true;
        });

        // entityBuilder.setExpand(true);
        entityBuilder.setMakeItemFn([this](const std::string & name, const Entity & entity)
        {
            return makeItem(name, entity);
        });
        entityBuilder.setUpdateItemFn([this](const std::string&, const Entity & entity, QTreeWidgetItem *  entityItem)
        {
            updateContainerItem(entity, entityItem);
            updateChildren(entity, entityItem);
            return true;
        });

        snapshotBuilder.setMakeItemFn([this](const armem::Time & time, const EntitySnapshot & snapshot)
        {
            QTreeWidgetItem* item = makeItem(toDateTimeMilliSeconds(time), snapshot);
            item->setData(int(Columns::KEY), Qt::ItemDataRole::UserRole, QVariant(static_cast<qlonglong>(snapshot.time().toMicroSeconds())));
            return item;
        });
        snapshotBuilder.setCompareFn([](const auto & pair, QTreeWidgetItem * item)
        {
            const EntitySnapshot& snapshot = pair.second;
            return armarx::detail::compare(static_cast<qlonglong>(snapshot.time().toMicroSeconds()),
                                           item->data(int(Columns::KEY), Qt::ItemDataRole::UserRole).toLongLong());
        });
        snapshotBuilder.setUpdateItemFn([this](const armem::Time&, const EntitySnapshot & snapshot, QTreeWidgetItem * snapshotItem)
        {
            updateContainerItem(snapshot, snapshotItem);
            updateChildren(snapshot, snapshotItem);
            return true;
        });

        instanceBuilder.setMakeItemFn([this](const EntityInstance & instance)
        {
            QTreeWidgetItem* item = makeItem("", instance);
            return item;
        });
        instanceBuilder.setCompareFn([](const EntityInstance & lhs, QTreeWidgetItem * rhsItem)
        {
            return armarx::detail::compare(lhs.index(), rhsItem->text(0).toInt());
        });
        instanceBuilder.setUpdateItemFn([this](const EntityInstance & instance, QTreeWidgetItem * instanceItem)
        {
            updateItemItem(instance, instanceItem);
            updateChildren(instance, instanceItem);
            return true;
        });

    }


    void TreeWidget::update(const armem::Memory& memory)
    {
        // Removing elements during the update can create unwanted signals triggering selection handling.
        handleSelections = false;
        updateChildren(memory, this);
        handleSelections = true;
        emit updated();
    }

    void TreeWidget::update(const std::map<std::string, const armem::Memory*>& memories)
    {
        handleSelections = false;
        updateChildren(memories, this);
        handleSelections = true;
        emit updated();
    }

    std::optional<MemoryID> TreeWidget::selectedID() const
    {
        return _selectedID;
    }

    void TreeWidget::handleSelection()
    {
        if (!handleSelections)
        {
            return;
        }
        QTreeWidgetItem* item = this->currentItem();
        if (!item)
        {
            return;
        }

        MemoryID id(item->data(int(Columns::ID), Qt::UserRole).toString().toStdString());

        // Has the selection changed due to an update? If yes, signal this.
        if (!_selectedID || id != *_selectedID)
        {
            _selectedID = id;

            const std::string levelName = item->data(int(Columns::LEVEL), Qt::UserRole).toString().toStdString();
            if (levelName == Memory().getLevelName())
            {
                emit memorySelected(*_selectedID);
            }
            else if (levelName == CoreSegment().getLevelName())
            {
                emit coreSegmentSelected(*_selectedID);
            }
            else if (levelName == ProviderSegment().getLevelName())
            {
                emit providerSegmentSelected(*_selectedID);
            }
            else if (levelName == Entity().getLevelName())
            {
                emit entitySelected(*_selectedID);
            }
            else if (levelName == EntitySnapshot().getLevelName())
            {
                emit snapshotSelected(*_selectedID);
            }
            else if (levelName == EntityInstance().getLevelName())
            {
                emit instanceSelected(*_selectedID);
            }
        }

        // Either selection or its contents have changed.
        emit selectedItemChanged(*_selectedID);
    }


    void TreeWidget::updateChildren(const armem::Memory& memory, QTreeWidget* tree)
    {
        updateChildren(std::map<std::string, const armem::Memory*> {{memory.name(), &memory}}, tree);
    }

    void TreeWidget::updateChildren(const std::map<std::string, const armem::Memory*>& memories, QTreeWidget* tree)
    {
        memoryBuilder.updateTree(tree, memories);
    }


    void TreeWidget::updateChildren(const armem::Memory& memory, QTreeWidgetItem* memoryItem)
    {
        coreSegmentBuilder.updateTree(memoryItem, memory.coreSegments);
    }

    void TreeWidget::updateChildren(const armem::CoreSegment& coreSeg, QTreeWidgetItem* coreSegItem)
    {
        provSegmentBuilder.updateTree(coreSegItem, coreSeg.providerSegments);
    }

    void TreeWidget::updateChildren(const armem::ProviderSegment& provSeg, QTreeWidgetItem* provSegItem)
    {
        entityBuilder.updateTree(provSegItem, provSeg.entities);
    }

    void TreeWidget::updateChildren(const armem::Entity& entity, QTreeWidgetItem* entityItem)
    {
        snapshotBuilder.updateTree(entityItem, entity.history);
    }

    void TreeWidget::updateChildren(const armem::EntitySnapshot& snapshot, QTreeWidgetItem* snapshotItem)
    {
        instanceBuilder.updateTree(snapshotItem, snapshot.instances);
    }

    void TreeWidget::updateChildren(const armem::EntityInstance& data, QTreeWidgetItem* dataItem)
    {
        (void) data, (void) dataItem;
    }

    QTreeWidgetItem* TreeWidget::makeItem(const std::string& key, const armem::detail::MemoryItem& memoryItem)
    {
        (void) key;
        QStringList columns;
        columns.insert(int(Columns::KEY), QString::fromStdString(memoryItem.getKeyString()));
        columns.insert(int(Columns::SIZE), "");
        columns.insert(int(Columns::TYPE), "");
        columns.insert(int(Columns::LEVEL), QString::fromStdString(simox::alg::capitalize_words(memoryItem.getLevelName())));
        columns.insert(int(Columns::ID), QString::fromStdString(memoryItem.id().str()));

        QTreeWidgetItem* item = new QTreeWidgetItem(columns);
        item->setData(int(Columns::LEVEL), Qt::UserRole, QString::fromStdString(memoryItem.getLevelName()));
        item->setData(int(Columns::ID), Qt::UserRole, QString::fromStdString(memoryItem.id().str()));
        item->setTextAlignment(int(Columns::SIZE), Qt::AlignRight);
        return item;
    }

    void TreeWidget::updateItemItem(const armem::detail::MemoryItem& level, QTreeWidgetItem* item)
    {
        (void) level, (void) item;
    }

    template <class... T>
    void TreeWidget::updateContainerItem(
        const armem::detail::MemoryContainer<T...>& container, QTreeWidgetItem* item)
    {
        updateItemItem(container, item);
        item->setText(int(Columns::SIZE), QString::number(container.size()));
    }

    template <class... T>
    void TreeWidget::updateTypedContainerItem(
        const armem::detail::TypedEntityContainer<T...>& container, QTreeWidgetItem* item)
    {
        updateContainerItem(container, item);
        std::string typeName;
        if (container.aronType())
        {
            typeName = container.aronType()->getName();

            std::string del = "::";
            size_t find = typeName.rfind(del);
            if (find != typeName.npos)
            {
                find += del.size();  // include del
                std::stringstream ss;
                ss << typeName.substr(find) << " (" << typeName.substr(0, find) << ")";
                typeName = ss.str();
            }
        }
        else
        {
            typeName = "(no Aron type)";
        }
        item->setText(int(Columns::TYPE), QString::fromStdString(typeName));
    }

}
