#include "MappingDataReader.h"

#include <vector>

#include <IceUtil/Time.h>

#include <SimoxUtility/algorithm/get_map_keys_values.h>

#include <ArmarXCore/core/logging/Logging.h>

#include <RobotAPI/libraries/armem/core/EntityInstance.h>
#include <RobotAPI/libraries/armem/core/EntitySnapshot.h>
#include <RobotAPI/libraries/armem/util/util.h>
#include <RobotAPI/libraries/armem_robot_mapping/aron/LaserScan.aron.generated.h>
#include <RobotAPI/libraries/armem_robot_mapping/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_mapping/types.h>
#include <RobotAPI/libraries/aron/core/codegenerator/codeWriter/cpp/AronCppClass.h>
#include "RobotAPI/libraries/armem_robot_localization/MemoryConnector.h"

namespace armarx::armem
{

    MappingDataReader::MappingDataReader(ManagedIceObject& component)
        : armarx::armem::MemoryConnector(component) {}

    MappingDataReader::~MappingDataReader() = default;

    void MappingDataReader::registerPropertyDefinitions(
        armarx::PropertyDefinitionsPtr& def)
    {
        ARMARX_DEBUG << "TransformReader: registerPropertyDefinitions";
        MemoryConnector::registerPropertyDefinitions(def);

        const std::string prefix = getPropertyPrefix();

        def->optional(properties.mappingMemoryName, prefix + "MappingMemoryName",
                      "Name of the mapping memory core segment to use.");

        def->optional(properties.memoryName, prefix + "MemoryName");
    }

    void MappingDataReader::connect()
    {
        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "TransformReader: Waiting for memory '"
                         << properties.memoryName << "' ...";
        auto result = useMemory(properties.memoryName);
        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;
            return;
        }

        ARMARX_IMPORTANT << "TransformReader: Connected to memory '"
                         << properties.memoryName;

        memoryReader.setReadingMemory(result.proxy);
    }

    armem::client::query::Builder
    MappingDataReader::buildQuery(const Query& query) const
    {
        armem::client::query::Builder qb;

        ARMARX_INFO << "Query for agent: " << query.agent
                    << " memory name: " << properties.mappingMemoryName;

        if (query.sensorList.empty()) // all sensors
        {
            // clang-format off
            qb
            .coreSegments().withName(properties.mappingMemoryName)
            .providerSegments().withName(query.agent)
            .entities().all()
            .snapshots().timeRange(
                IceUtil::Time::microSeconds(query.timeRange.min),
                IceUtil::Time::microSeconds(query.timeRange.max));
            // clang-format on
        }
        else
        {
            // clang-format off
            qb
            .coreSegments().withName(properties.mappingMemoryName)
            .providerSegments().withName(query.agent)
            .entities().withNames(query.sensorList)
            .snapshots().timeRange(
                IceUtil::Time::microSeconds(query.timeRange.min),
                IceUtil::Time::microSeconds(query.timeRange.max));
            // clang-format on
        }

        return qb;

    }

    MappingDataReader::Result
    MappingDataReader::queryData(const Query& query) const
    {
        const auto qb = buildQuery(query);

        ARMARX_IMPORTANT << "[MappingDataReader] query ... ";

        const armem::client::QueryResult qResult =
            memoryReader.query(qb.buildQueryInput());

        ARMARX_DEBUG << "[MappingDataReader] result: " << qResult;

        if (not qResult.success)
        {
            ARMARX_WARNING << "Failed to query data from memory: "
                           << qResult.errorMessage;
            return {.laserScans = {},
                    .sensors = {},
                    .status = Result::Status::Error,
                    .errorMessage = qResult.errorMessage};
        }

        // now create result from memory
        const auto& entities =
            qResult.memory.getCoreSegment(properties.mappingMemoryName)
            .getProviderSegment(query.agent)
            .entities;

        auto laserScans = transformAllOfType<aron::LaserScanStamped>(
                              entities,
                              [](const aron::LaserScanStamped & aronLaserScan) -> LaserScanStamped
        {
            LaserScanStamped laserScan;
            fromAron(aronLaserScan, laserScan);
            return laserScan;
        });

        const auto sensors = simox::alg::get_keys(entities);

        return {.laserScans = std::move(laserScans),
                .sensors = sensors,
                .status = Result::Status::Success,
                .errorMessage = ""};
    }

} // namespace armarx::armem
