/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "ObjectClass.h"


namespace armarx::aron::cppcodegenerator::serializer
{
    // constructors
    ObjectClassSerializer::ObjectClassSerializer(const typenavigator::ObjectNavigatorPtr& e) :
        Serializer(e->getObjectName(), simox::meta::get_type_name(typeid(data::AronDict)), simox::meta::get_type_name(typeid(type::AronObject))),
        navigator(e)
    {
        //AddObjectTypeToAllObjectTypesList(ObjectNavigatorPtr(this));
    }

    std::vector<CppFieldPtr> ObjectClassSerializer::getPublicVariableDeclarations(const std::string&) const
    {
        std::vector<CppFieldPtr> fields;
        for (const auto& [key, member] : navigator->getMemberTypes())
        {
            auto member_s = FromAronTypeNaviagtorPtr(member);
            std::vector<CppFieldPtr> member_fields = member_s->getPublicVariableDeclarations(key);
            fields.insert(fields.end(), member_fields.begin(), member_fields.end());
        }
        return fields;
    }

    std::vector<std::pair<std::string, std::string>> ObjectClassSerializer::getCtorInitializers(const std::string&) const
    {
        return {};
    }

    CppBlockPtr ObjectClassSerializer::getCtorBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getResetBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::reset();");
        }

        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getResetBlock(key);
            b->appendBlock(b2);
        }
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getInitializeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::initialize();");
        }

        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getInitializeBlock(key);
            b->appendBlock(b2);
        }
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getWriteInitialTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::writeInitialType(w);");
        }

        b->addLine("w.writeStartObject(\"" + navigator->getObjectName() + "\");");
        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            b->addLine("w.writeKey(\"" + key + "\");");
            CppBlockPtr b2 = child_s->getWriteInitialTypeBlock(child_s->getFullCppTypename());
            b->appendBlock(b2);
        }
        b->addLine("w.writeEndObject();");
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getWriteCurrentTypeBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::writeCurrentType(w);");
        }

        b->addLine("w.writeStartObject(\"" + navigator->getObjectName() + "\");");
        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            b->addLine("w.writeKey(\"" + key + "\");");
            CppBlockPtr b2 = child_s->getWriteCurrentTypeBlock(key);
            b->appendBlock(b2);
        }
        b->addLine("w.writeEndObject();");
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getWriteBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::write(w, aron_type);");
        }

        b->addLine("w.writeStartDict();");
        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            b->addLine("w.writeKey(\"" + key + "\");");
            CppBlockPtr b2 = child_s->getWriteBlock(key);
            b->appendBlock(b2);
        }
        b->addLine("w.writeEndDict();");
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getReadBlock(const std::string&) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine(extends_s->getFullCppTypename() + "::read(r);");
        }

        b->addLine("r.readStartDict();");
        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            const auto child_s = FromAronTypeNaviagtorPtr(child);
            b->addLine("r.loadMember(\"" + key + "\");");
            CppBlockPtr b2 = child_s->getReadBlock(key);
            b->appendBlock(b2);
        }
        b->addLine("r.readEndDict();");
        return b;
    }

    CppBlockPtr ObjectClassSerializer::getEqualsBlock(const std::string&, const std::string& otherInstanceAccessor) const
    {
        CppBlockPtr b = CppBlockPtr(new CppBlock());
        if (navigator->getExtends() != nullptr)
        {
            const auto extends_s = FromAronTypeNaviagtorPtr(navigator->getExtends());
            b->addLine("if (not (" + extends_s->getFullCppTypename() + "::operator== (" + otherInstanceAccessor + ")))");
            b->addLine("\t return false;");
        }
        for (const auto& [key, child] : navigator->getMemberTypes())
        {
            auto child_s = FromAronTypeNaviagtorPtr(child);
            CppBlockPtr b2 = child_s->getEqualsBlock(key, otherInstanceAccessor + "." + key);
            b->appendBlock(b2);
        }
        return b;
    }
}

