/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NlohmannJSONReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/data/primitive/Primitive.h>



namespace armarx::aron::dataIO::reader
{
    NlohmannJSONReader::NlohmannJSONReader(const nlohmann::json& n) :
        input(n)
    {
    }

    NlohmannJSONReader::NlohmannJSONReader(const std::string& n) :
        input(nlohmann::json::parse(n))
    {
    }

    nlohmann::json NlohmannJSONReader::getNextAndIncrease()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        NlohmannJSONReaderTokenPtr lastToken = stack.top();
        return lastToken->getNextElementAndIncreaseCnt();
    }

    nlohmann::json NlohmannJSONReader::getNext()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        NlohmannJSONReaderTokenPtr lastToken = stack.top();
        return lastToken->getNextElement();
    }

    int NlohmannJSONReader::readStartDict()
    {
        nlohmann::json current_json = getNextAndIncrease();
        int c = current_json.size();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(data::eDict, current_json);
        stack.push(newToken);
        return c;
    }

    bool NlohmannJSONReader::readEndDict()
    {
        auto token = stack.top();
        token->assertType(data::eDict);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NlohmannJSONReader::readStartList()
    {
        nlohmann::json current_json = getNextAndIncrease();
        int c = current_json.size();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(data::eList, current_json);
        stack.push(newToken);
        return c;
    }

    bool NlohmannJSONReader::readEndList()
    {
        auto token = stack.top();
        token->assertType(data::eList);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    // Complex Data
    std::tuple<std::vector<int>, std::string> NlohmannJSONReader::readStartNDArray()
    {
        nlohmann::json j = getNext();
        std::vector<int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        return {dims, type};
    }

    bool NlohmannJSONReader::readEndNDArray(unsigned char* data)
    {
        if (data == NULL)
        {
            throw error::AronException("NlohmannJSONReader", "readEndNDArray", "The corresponding data pointer of an complex aron ptr is NULL.");
        }
        nlohmann::json j = getNextAndIncrease();
        std::vector<int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::vector<unsigned char> d = j[io::Data::READER_WRITER_NDARRAY_DATA_SLUG];
        memcpy(data, d.data(), std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>()));
        return true;
    }

    // Read primitives
    void NlohmannJSONReader::readPrimitive(int& t)
    {
        nlohmann::json j = getNextAndIncrease();
        t = j;
    }

    void NlohmannJSONReader::readPrimitive(long& t)
    {
        nlohmann::json j = getNextAndIncrease();
        t = j;
    }

    void NlohmannJSONReader::readPrimitive(float& t)
    {
        nlohmann::json j = getNextAndIncrease();
        t = j;
    }

    void NlohmannJSONReader::readPrimitive(double& t)
    {
        nlohmann::json j = getNextAndIncrease();
        t = j;
    }

    void NlohmannJSONReader::readPrimitive(std::string& t)
    {
        nlohmann::json j = getNextAndIncrease();
        t = j;
    }

    void NlohmannJSONReader::readPrimitive(bool& t)
    {
        nlohmann::json j = getNextAndIncrease();
        t = j;
    }

    std::string NlohmannJSONReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NlohmannJSONReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    // Helper functions
    data::Descriptor NlohmannJSONReader::getTypeOfNext(const type::Descriptor hint) const
    {
        auto token = stack.top();
        return token->getTypeOfNext(hint);
    }
}
