/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller (fabian dot peller at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <numeric>

#include "NlohmannJSONWriter.h"

namespace armarx::aron::dataIO::writer
{

    // Containers
    void NlohmannJSONWriter::writeStartDict()
    {
        nlohmann::json data;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(data::eDict, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndDict()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    void NlohmannJSONWriter::writeStartList()
    {
        nlohmann::json data;
        auto new_token = std::make_shared<NlohmannJSONWriterToken>(data::eList, data);
        stack.push(new_token);
    }

    void NlohmannJSONWriter::writeEndList()
    {
        lastRemovedToken = stack.top();
        stack.pop();

        if (stack.size() > 0)
        {
            auto prevToken = stack.top();
            prevToken->addElement(lastRemovedToken->getElement());
        }
    }

    // Complex Types
    void NlohmannJSONWriter::writeNDArray(const std::vector<int>& dims, const std::string& t, const unsigned char* data) \
    {
        NlohmannJSONWriterTokenPtr token = stack.top();
        nlohmann::json j;
        j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG] = dims;
        j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG] = t;
        int elements = std::accumulate(std::begin(dims), std::end(dims), 1, std::multiplies<int>());
        std::vector<unsigned char> d = std::vector<unsigned char>(elements);
        memcpy(d.data(), data, elements);
        j[io::Data::READER_WRITER_NDARRAY_DATA_SLUG] = d;
        token->addElement(j);
    }

    // Primitive
    void NlohmannJSONWriter::writePrimitive(const int x)
    {
        auto token = stack.top();
        nlohmann::json j(x);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePrimitive(const long x)
    {
        auto token = stack.top();
        nlohmann::json j(x);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePrimitive(const float x)
    {
        auto token = stack.top();
        nlohmann::json j(x);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePrimitive(const double x)
    {
        auto token = stack.top();
        nlohmann::json j(x);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePrimitive(const std::string& x)
    {
        auto token = stack.top();
        nlohmann::json j(x);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writePrimitive(const bool x)
    {
        auto token = stack.top();
        nlohmann::json j(x);
        token->addElement(j);
    }

    void NlohmannJSONWriter::writeKey(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

}
