/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "NlohmannJSONReader.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/type/primitive/Primitive.h>



namespace armarx::aron::typeIO::reader
{
    NlohmannJSONReader::NlohmannJSONReader(const nlohmann::json& n) :
        input(n)
    {
    }

    NlohmannJSONReader::NlohmannJSONReader(const std::string& n) :
        input(n)
    {
    }

    nlohmann::json NlohmannJSONReader::getNextAndIncrease()
    {
        if (!readInitialStart)
        {
            readInitialStart = true;
            return input;
        }
        NlohmannJSONReaderTokenPtr lastToken = stack.top();
        return lastToken->getNextElementAndIncreaseCnt();
    }

    std::tuple<std::string, int> NlohmannJSONReader::readStartObject()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::eObject, current_json);
        stack.push(newToken);
        return {newToken->getElementName(), newToken->getElementChildrenSize()};
    }

    bool NlohmannJSONReader::readEndObject()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::eObject);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NlohmannJSONReader::readStartDict()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::eDict, current_json);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NlohmannJSONReader::readEndDict()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::eDict);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NlohmannJSONReader::readStartList()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::eList, current_json);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NlohmannJSONReader::readEndList()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::eList);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NlohmannJSONReader::readStartTuple()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::eTuple, current_json);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NlohmannJSONReader::readEndTuple()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::eTuple);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    int NlohmannJSONReader::readStartPair()
    {
        nlohmann::json current_json = getNextAndIncrease();
        auto newToken = std::make_shared<NlohmannJSONReaderToken>(type::ePair, current_json);
        stack.push(newToken);
        return newToken->getElementChildrenSize();
    }

    bool NlohmannJSONReader::readEndPair()
    {
        NlohmannJSONReaderTokenPtr token = stack.top();
        token->assertType(type::ePair);

        if (token->finishedElement())
        {
            stack.pop();
            return true;
        }
        return false;
    }

    std::tuple<std::vector<int>, std::string> NlohmannJSONReader::readEigenMatrix()
    {
        nlohmann::json j = getNextAndIncrease();
        std::vector<int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        return std::make_tuple(dims, type);
    }

    std::string NlohmannJSONReader::readEigenQuaternion()
    {
        nlohmann::json j = getNextAndIncrease();
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        return type;
    }

    std::tuple<unsigned int, unsigned int, std::string> NlohmannJSONReader::readIVTCByteImage()
    {
        nlohmann::json j = getNextAndIncrease();
        std::vector<int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        return std::make_tuple(dims[0], dims[1], type);
    }

    std::tuple<std::vector<int>, std::string> NlohmannJSONReader::readOpenCVMat()
    {
        nlohmann::json j = getNextAndIncrease();
        std::vector<int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        return std::make_tuple(dims, type);
    }

    std::tuple<unsigned int, unsigned int, std::string> NlohmannJSONReader::readPCLPointCloud()
    {
        nlohmann::json j = getNextAndIncrease();
        std::vector<int> dims = j[io::Data::READER_WRITER_NDARRAY_DIMENSIONS_SLUG];
        std::string type = j[io::Data::READER_WRITER_NDARRAY_TYPE_SLUG];
        return std::make_tuple(dims[0], dims[1], type);
    }

    void NlohmannJSONReader::readPosition()
    {
        getNextAndIncrease();
        return;
    }

    void NlohmannJSONReader::readOrientation()
    {
        getNextAndIncrease();
        return;
    }

    void NlohmannJSONReader::readPose()
    {
        getNextAndIncrease();
        return;
    }

    // Read primitives
    void NlohmannJSONReader::readInt()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    void NlohmannJSONReader::readLong()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    void NlohmannJSONReader::readFloat()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    void NlohmannJSONReader::readDouble()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    void NlohmannJSONReader::readString()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    void NlohmannJSONReader::readBool()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    void NlohmannJSONReader::readTime()
    {
        nlohmann::json j = getNextAndIncrease();
    }

    std::string NlohmannJSONReader::readKey()
    {
        auto token = stack.top();
        return token->getCurrentKey();
    }

    void NlohmannJSONReader::loadMember(const std::string& k)
    {
        auto token = stack.top();
        token->setCurrentKey(k);
    }

    // Helper functions
    type::Descriptor NlohmannJSONReader::getTypeOfNext() const
    {
        auto token = stack.top();
        return token->getTypeOfNext();
    }
}
