/*
* This file is part of ArmarX.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @author     Fabian Peller (fabian dot peller at kit dot edu)
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

// STD/STL
#include <memory>
#include <numeric>

// Header
#include "Visitor.h"

// ArmarX
#include <RobotAPI/libraries/aron/core/Resolver.h>



namespace armarx::aron::typeIO
{

    void Visitor::SetupWriterFromAronTypePtr(WriterInterface& writer, const typenavigator::NavigatorPtr& aron)
    {
        SetupWriterFromAronTypePtr(writer, aron->getResult());
    }

    void Visitor::SetupWriterFromAronTypePtr(WriterInterface& writer, const type::AronTypePtr& aron)
    {
        type::Descriptor desc = Resolver::GetDescriptor(aron);
        switch (desc)
        {
            case type::eObject:
            {
                type::AronObjectPtr casted = type::AronObjectPtr::dynamicCast(aron);
                writer.writeStartObject(casted->objectName);
                for (const auto& [key, value] : casted->elementTypes)
                {
                    writer.writeKey(key);
                    Visitor::SetupWriterFromAronTypePtr(writer, value);
                }
                writer.writeEndObject();
                break;
            }
            case type::eDict:
            {
                type::AronDictPtr casted = type::AronDictPtr::dynamicCast(aron);
                writer.writeStartDict();
                Visitor::SetupWriterFromAronTypePtr(writer, casted->acceptedType);
                writer.writeEndDict();
                break;
            }
            case type::eTuple:
            {
                type::AronTuplePtr casted = type::AronTuplePtr::dynamicCast(aron);
                writer.writeStartTuple();
                for (const auto& value : casted->elementTypes)
                {
                    Visitor::SetupWriterFromAronTypePtr(writer, value);
                }
                writer.writeEndTuple();
                break;
            }
            case type::eList:
            {
                type::AronListPtr casted = type::AronListPtr::dynamicCast(aron);
                writer.writeStartList();
                Visitor::SetupWriterFromAronTypePtr(writer, casted->acceptedType);
                writer.writeEndList();
                break;
            }
            case type::eEigenMatrix:
            {
                type::AronEigenMatrixPtr casted = type::AronEigenMatrixPtr::dynamicCast(aron);
                writer.writeEigenMatrix(casted->dimensions, casted->typeName);
                break;
            }
            case type::eEigenQuaternion:
            {
                type::AronEigenQuaternionPtr casted = type::AronEigenQuaternionPtr::dynamicCast(aron);
                writer.writeEigenQuaternion(casted->typeName);
                break;
            }
            case type::eIVTCByteImage:
            {
                type::AronIVTCByteImagePtr casted = type::AronIVTCByteImagePtr::dynamicCast(aron);
                writer.writeIVTCByteImage(casted->width, casted->height, casted->typeName);
                break;
            }
            case type::eOpenCVMat:
            {
                type::AronOpenCVMatPtr casted = type::AronOpenCVMatPtr::dynamicCast(aron);
                writer.writeOpenCVMat(casted->dimensions, casted->typeName);
                break;
            }
            case type::ePCLPointCloud:
            {
                type::AronPCLPointCloudPtr casted = type::AronPCLPointCloudPtr::dynamicCast(aron);
                writer.writePCLPointCloud(casted->width, casted->height, casted->typeName);
                break;
            }
            case type::ePosition:
            {
                writer.writePosition();
                break;
            }
            case type::eOrientation:
            {
                writer.writeOrientation();
                break;
            }
            case type::ePose:
            {
                writer.writePose();
                break;
            }
            case type::eInt:
            {
                type::AronIntPtr casted = type::AronIntPtr::dynamicCast(aron);
                writer.writeInt();
                break;
            }
            case type::eLong:
            {
                type::AronLongPtr casted = type::AronLongPtr::dynamicCast(aron);
                writer.writeLong();
                break;
            }
            case type::eFloat:
            {
                type::AronFloatPtr casted = type::AronFloatPtr::dynamicCast(aron);
                writer.writeFloat();
                break;
            }
            case type::eDouble:
            {
                type::AronDoublePtr casted = type::AronDoublePtr::dynamicCast(aron);
                writer.writeDouble();
                break;
            }
            case type::eString:
            {
                type::AronStringPtr casted = type::AronStringPtr::dynamicCast(aron);
                writer.writeString();
                break;
            }
            case type::eBool:
            {
                type::AronBoolPtr casted = type::AronBoolPtr::dynamicCast(aron);
                writer.writeBool();
                break;
            }
            case type::eTime:
            {
                type::AronTimePtr casted = type::AronTimePtr::dynamicCast(aron);
                writer.writeTime();
                break;
            }
            default:
            {
                throw error::DescriptorNotValidException("LegacyAronTypeWriter", "SetupWriterFromAronTypePtr", "Type-Type could not be resolved. The ice_id of the input was: " + aron->ice_id(), desc);
            }
        }
    }
}
