/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "NDArray.h"

// Simox
#include <SimoxUtility/algorithm/string.h>

// ArmarX
#include <RobotAPI/libraries/aron/core/navigator/data/NavigatorFactory.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/OpenCVMat.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/EigenMatrix.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/EigenQuaternion.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/IVTCByteImage.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/PCLPointCloud.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/Position.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/Orientation.h>
#include <RobotAPI/libraries/aron/core/navigator/type/ndarray/Pose.h>

namespace armarx::aron::datanavigator
{
    // constructors
    NDArrayNavigator::NDArrayNavigator(const Path& path) :
        aron::Navigator<data::Descriptor, data::AronData>::Navigator(data::Descriptor::eNDArray, path),
        Navigator(data::Descriptor::eNDArray, path),
        aron(new data::AronNDArray())
    {

    }

    NDArrayNavigator::NDArrayNavigator(const data::AronNDArrayPtr& o, const Path& path) :
        aron::Navigator<data::Descriptor, data::AronData>::Navigator(data::Descriptor::eNDArray, path),
        Navigator(data::Descriptor::eNDArray, path),
        aron(o)
    {
        CheckAronPtrForNull("NDArrayNavigator", "NDArrayNavigator", getPath(), aron);
    }

    NDArrayNavigator::NDArrayNavigator(const std::vector<int>& dim, const std::string& t, const std::vector<unsigned char>& data, const Path& path) :
        NDArrayNavigator(data::AronNDArrayPtr(new data::AronNDArray(dim, t, data)), path)
    {
    }

    // operators
    bool NDArrayNavigator::operator==(const NDArrayNavigator& other) const
    {
        const auto otherAron = other.toAronNDArrayPtr();
        if (aron->dimensions != otherAron->dimensions)
        {
            return false;
        }
        if (aron->type != otherAron->type)
        {
            return false;
        }
        // performs memcmp
        if (aron->data != otherAron->data)
        {
            return false;
        }
        return true;
    }

    bool NDArrayNavigator::equalsDataNavigator(const NavigatorPtr& other) const
    {
        if (other == nullptr)
        {
            return false;
        }
        NDArrayNavigatorPtr casted = DynamicCast(other);
        return equalsDeep(casted);
    }

    bool NDArrayNavigator::equalsDeep(const NDArrayNavigatorPtr& other) const
    {
        return *this == *other;
    }


    // static methods
    NDArrayNavigatorPtr NDArrayNavigator::DynamicCast(const NavigatorPtr& n)
    {
        NDArrayNavigatorPtr casted = std::dynamic_pointer_cast<NDArrayNavigator>(n);
        return casted;
    }

    NDArrayNavigator NDArrayNavigator::DynamicCast(Navigator& n)
    {
        return dynamic_cast<NDArrayNavigator&>(n);
    }

    NDArrayNavigatorPtr NDArrayNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckDataNavigatorPtrForNull("NDArrayNavigator", "DynamicCastAndCheck[Before]", n);
        NDArrayNavigatorPtr casted = NDArrayNavigator::DynamicCast(n);
        CheckDataNavigatorPtrForNull("NDArrayNavigator", "DynamicCastAndCheck[After]", n->getPath(), casted);
        return casted;
    }

    NDArrayNavigatorPtr NDArrayNavigator::FromAronNDArrayPtr(const data::AronNDArrayPtr& aron)
    {
        return std::make_shared<NDArrayNavigator>(aron);
    }

    data::AronNDArrayPtr NDArrayNavigator::ToAronNDArrayPtr(const NDArrayNavigatorPtr& navigator)
    {
        return navigator ? navigator->toAronNDArrayPtr() : nullptr;
    }

    // public member functions
    unsigned char* NDArrayNavigator::getData() const
    {
        return aron->data.data();
    }

    void NDArrayNavigator::setData(unsigned int elements, const unsigned char* src)
    {
        aron->data = std::vector<unsigned char>(elements);
        memcpy(aron->data.data(), src, elements);
    }

    std::vector<int> NDArrayNavigator::getDimensions() const
    {
        return aron->dimensions;
    }

    void NDArrayNavigator::setDimensions(const std::vector<int>& d)
    {
        aron->dimensions = d;
    }

    void NDArrayNavigator::addDimension(int i)
    {
        aron->dimensions.push_back(i);
    }

    std::string NDArrayNavigator::getType() const
    {
        return aron->type;
    }

    void NDArrayNavigator::setType(const std::string& t)
    {
        if (t.empty())
        {
            throw error::AronException("NDArrayNavigator", "setType", "The type cannot be empty", getPath());
        }

        aron->type = t;
    }

    data::AronNDArrayPtr NDArrayNavigator::toAronNDArrayPtr() const
    {
        return aron;
    }

    // virtual implementations
    data::AronDataPtr NDArrayNavigator::getResult() const
    {
        return toAronNDArrayPtr();
    }

    std::string NDArrayNavigator::getName() const
    {
        return "AronNDArray<" + simox::alg::to_string(aron->dimensions, ", ") + ", " + aron->type + ">";
    }

    typenavigator::NavigatorPtr NDArrayNavigator::recalculateType() const
    {
        // We set all to openCVMat because it accepts n-dimensional types
        typenavigator::OpenCVMatNavigatorPtr typenav = typenavigator::OpenCVMatNavigatorPtr(new typenavigator::OpenCVMatNavigator(getPath()));
        typenav->setDimensions(std::vector<int>(aron->dimensions.begin(), std::prev(aron->dimensions.end())));
        typenav->setTypename(aron->type);
        return typenav;
    }

    bool NDArrayNavigator::fullfillsType(const typenavigator::NavigatorPtr& type) const
    {
        if (!Resolver::Correspond(type->getDescriptor(), getDescriptor()))
        {
            return false;
        }

        switch (type->getDescriptor())
        {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
case type::e##upperType: \
{ \
    typenavigator::upperType##NavigatorPtr casted = typenavigator::upperType##Navigator::DynamicCast(type); \
    if (std::vector<int>(aron->dimensions.begin(), std::prev(aron->dimensions.end())) != casted->getDimensions() || aron->type != casted->getTypename()) \
    { \
        return false; \
    } \
    break; \
}

            HANDLE_NDARRAY_TYPES
#undef RUN_ARON_MACRO
            default:
                throw error::AronException("NDArrayNavigator", "fullfillsType", "Could not cast a type to an NDArray-Type", getPath());
        }
        return true;
    }

    std::string NDArrayNavigator::DimensionsToString(const std::vector<int>& dimensions)
    {
        std::stringstream ss;
        ss << "(" << simox::alg::join(simox::alg::multi_to_string(dimensions), ", ") << ")";
        return ss.str();
    }

    std::vector<NavigatorPtr> NDArrayNavigator::getChildren() const
    {
        return {};
    }

    size_t NDArrayNavigator::childrenSize() const
    {
        return 0;
    }
}
